USE master;
GO

CREATE PROCEDURE sp_FixOrphanedUsers @test BIT = 0, @all BIT = 0
/*
  Created by: Jen McCon, Jen@MidnightDBA.com, April 2014

  Purpose: 
		Repair all orphaned users in a database.

  Instructions:
		This procedure name MUST begin with "sp_", and must be created in the master database.
		Then, run this procedure from any database to repair orphaned users in that database.

  Parameters:
		@test - This will run the repair, add a dummy orphan, AND display code to help with troubleshooting logins.
		@all -  This parameter is used in conjunction with "sp_FixOrphanedUsersAllDB"; it causes data
				to be added to a global temporary table.

  Reference:
		See the article "Applied SQL: Orphan a user, fix orphaned users" (published 5/5/2014) on 
		http://www.MidnightDBA.com/Jen for a test scenario; you can use that script to create 
		a database and generate an orphaned user, then test this procedure	on it.

  Example execution:
		EXEC Demo..sp_FixOrphanedUsers;				-- This fixes orphaned users in the Demo database.
		EXEC Demo..sp_FixOrphanedUsers @test = 1;	-- This fixes orphaned users in the Demo DB, and displays extra code.

*/
AS 
	---- Setup ----
    DECLARE @i INT ,
        @userName SYSNAME ,
        @cmd NVARCHAR(4000) ,
        @err INT ,
        @msg NVARCHAR(2000);

    CREATE TABLE #LoginReport
        (
          ID INT IDENTITY(1, 1) ,
          UserName SYSNAME ,
          userSID VARBINARY(85) ,
          Outcome NVARCHAR(1000) NULL
        );

    INSERT  INTO #LoginReport
            ( UserName, userSID )
            EXEC sp_change_users_login 'Report';
			
    IF @test = 1 
        INSERT  INTO #LoginReport
                ( UserName, userSID )
        VALUES  ( 'testuser', NULL ); 
	
	
---- Loop through orphaned users and repair ---- 
    SELECT  @i = MAX(ID)
    FROM    #LoginReport;	

    WHILE @i > 0 
        BEGIN
            SELECT  @userName = UserName
            FROM    #LoginReport
            WHERE   ID = @i;

            -- Build the sp_change_users_login repair command
            SET @cmd = N'EXEC [' + DB_NAME()
                + ']..sp_Change_Users_Login ''AUTO_FIX'', ''' + @userName
                + ''';';

            PRINT ( N'---- Repairing orphaned user: ' + @userName + ' ----' );

            IF @test = 1 
                PRINT @cmd;		
            
			------------- BEGIN TRY/CATCH -------------
            BEGIN TRY
           
                EXEC sp_executesql @cmd;		-- Run the user repair command.
				
                UPDATE  #LoginReport
                SET     Outcome = 'Repaired'
                WHERE   ID = @i;

            END TRY
            BEGIN CATCH
                SELECT  @err = @@error ,
                        @msg = ERROR_MESSAGE();

                IF @err != 0 
                    BEGIN
                        RAISERROR('ERROR: Could not repair the login. Continuing.', 10, 1); 

                        IF NOT EXISTS ( SELECT  *
                                        FROM    sys.server_principals
                                        WHERE   name = @username ) 
                            UPDATE  #LoginReport
                            SET     Outcome = 'Error. This user does not have an existing login of the same name; it cannot be repaierd automatically.'
                            WHERE   ID = @i;
                        ELSE 
                            UPDATE  #LoginReport
                            SET     Outcome = 'Error (unknown cause). This user needs to be repaired manually.'
                            WHERE   ID = @i;
                    END

            END CATCH
			-------------- END TRY/CATCH --------------
			
            SET @i = @i - 1;

        END

    -- If @all = 0, we only want to run this for a single database. Just return the operation outcome.
	IF @all = 0 
        SELECT  UserName ,
                Outcome
        FROM    #LoginReport
        ORDER BY Outcome ,
                UserName;
    
	-- If @all = 1, we want to run this for all databases. Insert data to a global temp table
	-- for use by the calling SP.
	ELSE IF @all = 1
        BEGIN 
            IF NOT EXISTS ( SELECT  *
                            FROM    tempdb.sys.tables
                            WHERE   name = '##AllDBReport' ) 
                CREATE TABLE ##AllDBReport
                    (
                      DBName SYSNAME ,
                      UserName SYSNAME ,
                      Outcome NVARCHAR(1000) NULL
                    );


            INSERT  INTO ##AllDBReport
                    ( DBName ,
                      UserName ,
                      Outcome 
                    )
                    SELECT  DB_NAME() AS DBName ,
                            UserName ,
                            Outcome
                    FROM    #LoginReport
                    ORDER BY Outcome ,
                            UserName;
        END		-- End of IF @all = 1.

    DROP TABLE #LoginReport;

GO




CREATE PROCEDURE sp_FixOrphanedUsersAllDB
/*
  Created by: Jen McCon, Jen@MidnightDBA.com, April 2014

  Purpose: 
	Run sp_FixOrphanedUsers for all databases, to repair all orphaned users.

  Instructions:
	This procedure name MUST begin with "sp_", and must be created in the master database.

  Reference:
	See the article "Applied SQL: Orphan a user, fix orphaned users" (published 5/5/2014) on 
	http://www.MidnightDBA.com/Jen for a test scenario; you can use that script to create 
	a database and generate an orphaned user, then test this procedure	on it.

  Example execution:
	EXEC sp_FixOrphanedUsersAllDB;

*/
AS 
    EXEC sp_msforeachdb 'USE [?]; 
	exec sp_executesql N''EXEC [?]..sp_FixOrphanedUsers @all=1;''';
	
    SELECT  *
    FROM    ##AllDBReport
    ORDER BY DBName ,
            Outcome ,
            UserName;

    DROP TABLE ##AllDBReport;

GO

